"""

Protocol AST declaration.

"""

from .protocol_ast import *


FRAME_ID = b"COZ\x03RE\x01"
MIN_FRAME_SIZE = len(FRAME_ID) + 1 + 2 + 2 + 2

FIRST_ROBOT_PACKET_ID = 0xb0

FIRMWARE_VERSION = 2381


BODY_COLOR = Enum("BodyColor", members=[
    EnumMember("UNKNOWN", -1),
    EnumMember("WHITE_v10", 0),
    EnumMember("RESERVED", 1),
    EnumMember("WHITE_v15", 2),
    EnumMember("CE_LM_v15", 3),
    EnumMember("LE_BL_v16", 4),
])
NV_ENTRY_TAG = Enum("NvEntryTag", base=16, members=[
    EnumMember("NVEntry_Invalid", 0xffffffff),
    EnumMember("NVEntry_GameSkillLevels", 0x180000),
    EnumMember("NVEntry_OnboardingData", 0x181000),
    EnumMember("NVEntry_GameUnlocks", 0x182000),
    EnumMember("NVEntry_FaceEnrollData", 0x183000),
    EnumMember("NVEntry_FaceAlbumData", 0x184000),
    EnumMember("NVEntry_NurtureGameData", 0x194000),
    EnumMember("NVEntry_InventoryData", 0x195000),
    EnumMember("NVEntry_LabAssignments", 0x196000),
    EnumMember("NVEntry_SavedCubeIDs", 0x197000),
    EnumMember("NVEntry_NEXT_SLOT", 0x198000),
    EnumMember("NVEntry_FACTORY_RESERVED1", 0x1c0000),
    EnumMember("NVEntry_FACTORY_RESERVED2", 0x1de000),
    EnumMember("NVEntry_BirthCertificate", 0x80000000),
    EnumMember("NVEntry_CameraCalib", 0x80000001),
    EnumMember("NVEntry_ToolCodeInfo", 0x80000002),
    EnumMember("NVEntry_CalibPose", 0x80000003),
    EnumMember("NVEntry_CalibMetaInfo", 0x80000004),
    EnumMember("NVEntry_ObservedCubePose", 0x80000005),
    EnumMember("NVEntry_IMUInfo", 0x80000006),
    EnumMember("NVEntry_CliffValOnDrop", 0x80000007),
    EnumMember("NVEntry_CliffValOnGround", 0x80000008),
    EnumMember("NVEntry_PlaypenTestResults", 0x80000010),
    EnumMember("NVEntry_FactoryLock", 0x80000011),
    EnumMember("NVEntry_VersionMagic", 0x80000012),
    EnumMember("NVEntry_CalibImage1", 0x80010000),
    EnumMember("NVEntry_CalibImage2", 0x80020000),
    EnumMember("NVEntry_CalibImage3", 0x80030000),
    EnumMember("NVEntry_CalibImage4", 0x80040000),
    EnumMember("NVEntry_CalibImage5", 0x80050000),
    EnumMember("NVEntry_CalibImage6", 0x80060000),
    EnumMember("NVEntry_ToolCodeImageLeft", 0x80100000),
    EnumMember("NVEntry_ToolCodeImageRight", 0x80110000),
    EnumMember("NVEntry_PrePlaypenResults", 0xc0000000),
    EnumMember("NVEntry_PrePlaypenCentroids", 0xc0000001),
    EnumMember("NVEntry_IMUAverages", 0xc0000004),
    EnumMember("NVEntry_FactoryBaseTag", 0xde000),
    EnumMember("NVEntry_FactoryBaseTagWithBCOffset", 0xde030),
])
NV_OPERATION = Enum("NvOperation", members=[
    EnumMember("NVOP_READ", 0),
    EnumMember("NVOP_WRITE", 1),
    EnumMember("NVOP_ERASE", 2),
    EnumMember("NVOP_WIPEALL", 3),
])
NV_RESULT = Enum("NvResult", members=[
    EnumMember("NV_OKAY", 0),
    EnumMember("NV_SCHEDULED", 1),
    EnumMember("NV_NO_DO", 2),
    EnumMember("NV_MORE", 3),
    EnumMember("NV_UNKNOWN_4", 4),
    EnumMember("NV_UNKNOWN_5", 5),
    EnumMember("NV_UNKNOWN_6", 6),
    EnumMember("NV_UNKNOWN_7", 7),
    EnumMember("NV_UNKNOWN_8", 8),
    EnumMember("NV_NOT_FOUND", -1),
    EnumMember("NV_NO_ROOM", -2),
    EnumMember("NV_ERROR", -3),
    EnumMember("NV_TIMEOUT", -4),
    EnumMember("NV_BUSY", -5),
    EnumMember("NV_BAD_ARGS", -6),
    EnumMember("NV_NO_MEM", -7),
    EnumMember("NV_LOOP", -8),
    EnumMember("NV_CORRUPT", -9),
])
UP_AXIS = Enum("UpAxis", members=[
    EnumMember("XNegative", 0),
    EnumMember("XPositive", 1),
    EnumMember("YNegative", 2),
    EnumMember("YPositive", 3),
    EnumMember("ZNegative", 4),
    EnumMember("ZPositive", 5),
    EnumMember("NumAxes", 6),
    EnumMember("UnknownAxis", 7),
])
OBJECT_TYPE = Enum("ObjectType", members=[
    EnumMember("InvalidObject", -1),
    EnumMember("UnknownObject", 0),
    EnumMember("Block_LIGHTCUBE1", 1),
    EnumMember("Block_LIGHTCUBE2", 2),
    EnumMember("Block_LIGHTCUBE3", 3),
    EnumMember("Block_LIGHTCUBE_GHOST", 4),
    EnumMember("FlatMat_GEARS_4x4", 5),
    EnumMember("FlatMat_LETTERS_4x4", 6),
    EnumMember("FlatMat_ANKI_LOGO_8BIT", 7),
    EnumMember("FlatMat_LAVA_PLAYTEST", 8),
    EnumMember("Platform_LARGE", 9),
    EnumMember("Bridge_LONG", 10),
    EnumMember("Bridge_SHORT", 11),
    EnumMember("Ramp_Basic", 12),
    EnumMember("Charger_Basic", 13),
    EnumMember("ProxObstacle", 14),
    EnumMember("CliffDetection", 15),
    EnumMember("CollisionObstacle", 16),
    EnumMember("CustomType00", 17),
    EnumMember("CustomType01", 18),
    EnumMember("CustomType02", 19),
    EnumMember("CustomType03", 20),
    EnumMember("CustomType04", 21),
    EnumMember("CustomType05", 22),
    EnumMember("CustomType06", 23),
    EnumMember("CustomType07", 24),
    EnumMember("CustomType08", 25),
    EnumMember("CustomType09", 26),
    EnumMember("CustomType10", 27),
    EnumMember("CustomType11", 28),
    EnumMember("CustomType12", 29),
    EnumMember("CustomType13", 30),
    EnumMember("CustomType14", 31),
    EnumMember("CustomType15", 32),
    EnumMember("CustomType16", 33),
    EnumMember("CustomType17", 34),
    EnumMember("CustomType18", 35),
    EnumMember("CustomType19", 36),
    EnumMember("CustomFixedObstacle", 37),
])
IMAGE_ENCODING = Enum("ImageEncoding", members=[
    EnumMember("NoneImageEncoding", 0),
    EnumMember("RawGray", 1),
    EnumMember("RawRGB", 2),
    EnumMember("YUYV", 3),
    EnumMember("BAYER", 4),
    EnumMember("JPEGGray", 5),
    EnumMember("JPEGColor", 6),
    EnumMember("JPEGColorHalfWidth", 7),
    EnumMember("JPEGMinimizedGray", 8),
    EnumMember("JPEGMinimizedColor", 9),
])
IMAGE_RESOLUTION = Enum("ImageResolution", members=[
    EnumMember("VerificationSnapshot", 0),
    EnumMember("QQQQVGA", 1),
    EnumMember("QQQVGA", 2),
    EnumMember("QQVGA", 3),
    EnumMember("QVGA", 4),
    EnumMember("CVGA", 5),
    EnumMember("VGA", 6),
    EnumMember("SVGA", 7),
    EnumMember("XGA", 8),
    EnumMember("SXGA", 9),
    EnumMember("UXGA", 10),
    EnumMember("QXGA", 11),
    EnumMember("QUXGA", 12),
    EnumMember("ImageResolutionCount", 13),
    EnumMember("ImageResolutionNone", 14),
])
IMAGE_SEND_MODE = Enum("ImageSendMode", members=[
    EnumMember("Off", 0),
    EnumMember("Stream", 1),
    EnumMember("SingleShot", 2),
])
DEBUG_DATA_ID = Enum("DebugDataID", base=16, members=[
    EnumMember("MAC_ADDRESS", 0x1572),              # 624
])
MOTOR_ID = Enum("MotorID", members=[
    EnumMember("MOTOR_LEFT_WHEEL", 0),
    EnumMember("MOTOR_RIGHT_WHEEL", 1),
    EnumMember("MOTOR_LIFT", 2),
    EnumMember("MOTOR_HEAD", 3),
])
PATH_EVENT_TYPE = Enum("PathEventType", members=[
    EnumMember("PATH_STARTED", 0),
    EnumMember("PATH_INTERRUPTED", 1),
    EnumMember("PATH_COMPLETED", 2),
])

LIGHT_STATE = Struct("LightState", arguments=[
    UInt16Argument("on_color"),
    UInt16Argument("off_color"),
    UInt8Argument("on_frames"),
    UInt8Argument("off_frames"),
    UInt8Argument("transition_on_frames"),
    UInt8Argument("transition_off_frames"),
    Int16Argument("offset"),
])
PATH_SEGMENT_SPEED = Struct("PathSegmentSpeed", arguments=[
    FloatArgument("speed_mmps"),
    FloatArgument("accel_mmps2"),
    FloatArgument("decel_mmps2"),
])


PROTOCOL = Protocol(

    enums=[
        BODY_COLOR,
        NV_ENTRY_TAG,
        NV_OPERATION,
        NV_RESULT,
        UP_AXIS,
        OBJECT_TYPE,
        IMAGE_ENCODING,
        IMAGE_RESOLUTION,
        IMAGE_SEND_MODE,
        DEBUG_DATA_ID,
        MOTOR_ID,
        PATH_EVENT_TYPE,
    ],

    structs=[
        LIGHT_STATE,
        PATH_SEGMENT_SPEED,
    ],

    packets=[
        Connect(),
        Disconnect(),
        Ping(),
        Keyframe(),

        Command(0x03, "LightStateCenter", group="lights", arguments=[
            FArrayArgument("states", data_type=LIGHT_STATE, length=3),      # top, middle, bottom
            UInt8Argument("unknown"),
        ]),
        Command(0x04, "CubeLights", group="objects", arguments=[
            FArrayArgument("states", data_type=LIGHT_STATE, length=4)
        ]),
        Command(0x05, "ObjectConnect", group="objects", arguments=[
            UInt32Argument("factory_id"),
            BoolArgument("connect"),
        ]),
        Command(0x08, "StreamObjectAccel", group="objects", arguments=[
            UInt32Argument("object_id"),
            BoolArgument("enable"),
        ]),
        Command(0x0a, "SetAccessoryDiscovery", group="objects", arguments=[
            BoolArgument("enable")
        ]),
        Command(0x0b, "SetHeadLight", group="camera", arguments=[
            BoolArgument("enable")
        ]),
        Command(0x10, "CubeId", group="objects", arguments=[
            UInt32Argument("object_id"),
            UInt8Argument("rotation_period_frames"),
        ]),
        Command(0x11, "LightStateSide", group="lights", arguments=[
            FArrayArgument("states", data_type=LIGHT_STATE, length=2),      # left, right
            UInt8Argument("unknown"),
        ]),
        Command(0x25, "Enable", group="system"),
        Command(0x32, "DriveWheels", group="motors", arguments=[
            FloatArgument("lwheel_speed_mmps"),
            FloatArgument("rwheel_speed_mmps"),
            FloatArgument("lwheel_accel_mmps2"),
            FloatArgument("rwheel_accel_mmps2"),
        ]),
        Command(0x33, "TurnInPlaceAtSpeed", group="motors", arguments=[
            FloatArgument("wheel_speed_mmps"),
            FloatArgument("wheel_accel_mmps2"),
            Int16Argument("direction"),
        ]),
        Command(0x34, "MoveLift", group="motors", arguments=[
            FloatArgument("speed_rad_per_sec"),
        ]),
        Command(0x35, "MoveHead", group="motors", arguments=[
            FloatArgument("speed_rad_per_sec"),
        ]),
        Command(0x36, "SetLiftHeight", group="motors", arguments=[
            FloatArgument("height_mm"),
            FloatArgument("max_speed_rad_per_sec", default=3.0),
            FloatArgument("accel_rad_per_sec2", default=20.0),
            FloatArgument("duration_sec"),
            UInt8Argument("action_id"),
        ]),
        Command(0x37, "SetHeadAngle", group="motors", arguments=[
            FloatArgument("angle_rad"),
            FloatArgument("max_speed_rad_per_sec", default=15.0),
            FloatArgument("accel_rad_per_sec2", default=20.0),
            FloatArgument("duration_sec"),
            UInt8Argument("action_id"),
        ]),
        Command(0x39, "TurnInPlace", group="motors", arguments=[
            FloatArgument("angle_rad"),
            FloatArgument("speed_rad_per_sec"),
            FloatArgument("accel_rad_per_sec2"),
            FloatArgument("angle_tolerance_rad"),
            UInt8Argument("unknown4"),
            UInt8Argument("unknown5"),
            BoolArgument("is_absolute"),
            UInt8Argument("action_id"),
        ]),
        Command(0x3b, "StopAllMotors", group="motors"),
        Command(0x3c, "ClearPath", group="motors", arguments=[
            UInt16Argument("unknown"),
        ]),
        Command(0x3d, "AppendPathSegLine", group="motors", arguments=[
            FloatArgument("from_x"),
            FloatArgument("from_y"),
            FloatArgument("to_x"),
            FloatArgument("to_y"),
            FloatArgument("speed_mmps"),        # PathSegmentSpeed
            FloatArgument("accel_mmps2"),
            FloatArgument("decel_mmps2"),
        ]),
        Command(0x3e, "AppendPathSegArc", group="motors", arguments=[
            FloatArgument("center_x"),
            FloatArgument("center_y"),
            FloatArgument("radius_mm"),
            FloatArgument("start_angle_rad"),
            FloatArgument("sweep_rad"),
            FloatArgument("speed_mmps"),        # PathSegmentSpeed
            FloatArgument("accel_mmps2"),
            FloatArgument("decel_mmps2"),
        ]),
        Command(0x3f, "AppendPathSegPointTurn", group="motors", arguments=[
            FloatArgument("x"),
            FloatArgument("y"),
            FloatArgument("angle_rad"),
            FloatArgument("angle_tolerance_rad"),
            FloatArgument("speed_mmps"),        # PathSegmentSpeed
            FloatArgument("accel_mmps2"),
            FloatArgument("decel_mmps2"),
            BoolArgument("unknown")
        ]),
        Command(0x40, "TrimPath", group="motors", arguments=[
            UInt8Argument("head"),
            UInt8Argument("tail"),
        ]),
        Command(0x41, "ExecutePath", group="motors", arguments=[
            UInt16Argument("event_id"),
            BoolArgument("unknown"),
        ]),
        Command(0x45, "SetOrigin", group="localization", arguments=[
            UInt32Argument("unknown0"),
            UInt32Argument("pose_frame_id"),
            UInt32Argument("pose_origin_id", default=1),
            FloatArgument("pose_x"),
            FloatArgument("pose_y"),
            UInt32Argument("unknown5", default=0x80000000),
        ]),
        Command(0x4b, "SyncTime", group="system", arguments=[
            UInt32Argument("timestamp"),
            UInt32Argument("unknown"),
        ]),
        Command(0x4c, "EnableCamera", group="camera", arguments=[
            EnumArgument("image_send_mode", IMAGE_SEND_MODE, default=1),
            EnumArgument("image_resolution", IMAGE_RESOLUTION, default=4),
        ]),
        Command(0x57, "SetCameraParams", group="camera", arguments=[
            FloatArgument("gain"),
            UInt16Argument("exposure_ms"),
            BoolArgument("auto_exposure_enabled"),
        ]),
        Command(0x58, "StartMotorCalibration", group="motors", arguments=[
            BoolArgument("head"),
            BoolArgument("lift"),
        ]),
        Command(0x60, "EnableStopOnCliff", group="motors", arguments=[
            BoolArgument("enable"),
        ]),
        Command(0x64, "SetRobotVolume", group="audio", arguments=[
            UInt16Argument("level"),
        ]),
        Command(0x66, "EnableColorImages", group="camera", arguments=[
            BoolArgument("enable"),
        ]),
        Command(0x81, "NvStorageOp", group="nv", arguments=[
            EnumArgument("tag", NV_ENTRY_TAG, data_type=UInt32Argument(), default=0xffffffff),
            Int32Argument("length"),
            EnumArgument("op", NV_OPERATION, data_type=UInt8Argument()),
            UInt8Argument("unknown"),
            VArrayArgument("data"),
        ]),
        Command(0x8d, "AbortAnimation", group="anim"),
        Command(0x8e, "OutputAudio", group="audio", arguments=[
            FArrayArgument("samples", length=744),
        ]),
        Command(0x8f, "NextFrame", group="display"),
        Command(0x91, "RecordHeading", group="anim"),
        Command(0x92, "TurnToRecordedHeading", group="anim"),
        Command(0x93, "AnimHead", group="anim", arguments=[
            UInt8Argument("duration_ms"),
            Int8Argument("variability_deg"),
            Int8Argument("angle_deg"),
        ]),
        Command(0x94, "AnimLift", group="anim", arguments=[
            UInt8Argument("duration_ms"),
            UInt8Argument("variability_mm"),
            UInt8Argument("height_mm"),
        ]),
        Command(0x97, "DisplayImage", group="display", arguments=[
            VArrayArgument("image"),
        ]),
        Command(0x98, "AnimBackpackLights", group="anim", arguments=[
            FArrayArgument("colors", data_type=UInt16Argument(), length=5),  # left, front, middle, back, right
        ]),
        Command(0x99, "AnimBody", group="anim", arguments=[
            Int16Argument("speed"),
            Int16Argument("unknown"),
        ]),
        Command(0x9a, "EndAnimation", group="anim"),
        Command(0x9b, "StartAnimation", group="anim", arguments=[
            UInt8Argument("anim_id")
        ]),
        Command(0x9f, "EnableAnimationState", group="system"),
        Command(0xa9, "ShutdownRobot", group="system"),
        Command(0xae, "WifiOff", group="system", arguments=[
            BoolArgument("enable"),
        ]),
        Command(0xaf, "FirmwareUpdate", group="firmware", arguments=[
            UInt16Argument("chunk_id"),
            FArrayArgument("data", length=1024)
        ]),

        Command(0xb0, "DebugData", group="debug", arguments=[
            UInt16Argument("debug_id"),             # See DebugDataID
            UInt16Argument("unused"),               # Always 0
            UInt16Argument("unknown2"),             # Source?
            Int8Argument("unknown3"),               # Level? Observed: -1, 1, 2, 3, 5
            VArrayArgument("data", data_type=UInt32Argument(), length_type=UInt8Argument())
        ]),
        Command(0xb4, "ObjectMoved", group="objects", arguments=[
            UInt32Argument("timestamp"),
            UInt32Argument("object_id"),
            FloatArgument("active_accel_x"),
            FloatArgument("active_accel_y"),
            FloatArgument("active_accel_z"),
            EnumArgument("axis_of_accel", UP_AXIS, data_type=UInt8Argument(), default=7),
        ]),
        Command(0xb5, "ObjectStoppedMoving", group="objects", arguments=[
            UInt32Argument("timestamp"),
            UInt32Argument("object_id"),
        ]),
        Command(0xb6, "ObjectTapped", group="objects", arguments=[
            UInt32Argument("timestamp"),
            UInt32Argument("object_id"),
            UInt8Argument("num_taps"),
            UInt8Argument("tap_time"),
            Int8Argument("tap_neg"),
            Int8Argument("tap_pos"),
        ]),
        Command(0xb9, "ObjectTapFiltered", group="objects", arguments=[
            UInt32Argument("timestamp"),
            UInt32Argument("object_id"),
            UInt8Argument("time"),
            UInt8Argument("intensity"),
        ]),
        Command(0xc4, "AcknowledgeAction", group="motors", arguments=[
            UInt8Argument("action_id"),
        ]),
        Command(0xc2, "RobotDelocalized", group="localization"),
        Command(0xc3, "RobotPoked", group="localization"),
        Command(0xc6, "PathFollowingEvent", group="motors", arguments=[
            UInt16Argument("event_id"),
            EnumArgument("event_type", PATH_EVENT_TYPE, data_type=UInt8Argument()),
        ]),
        Command(0xc9, "HardwareInfo", group="system", arguments=[
            UInt32Argument("serial_number_head"),
            UInt8Argument("unknown1"),
            UInt8Argument("unknown2"),
        ]),
        Command(0xca, "AnimationStarted", group="anim", arguments=[
            UInt8Argument("anim_id")
        ]),
        Command(0xcb, "AnimationEnded", group="anim", arguments=[
            UInt8Argument("anim_id")
        ]),
        Command(0xcd, "NvStorageOpResult", group="nv", arguments=[
            EnumArgument("tag", NV_ENTRY_TAG, data_type=UInt32Argument(), default=0xffffffff),
            Int32Argument("length"),
            EnumArgument("op", NV_OPERATION, data_type=UInt8Argument()),
            EnumArgument("result", NV_RESULT, data_type=Int8Argument()),
            VArrayArgument("data"),
        ]),
        Command(0xce, "ObjectPowerLevel", group="objects", arguments=[
            UInt32Argument("object_id"),
            UInt32Argument("missed_packets"),
            UInt8Argument("battery_level"),
        ]),
        Command(0xd0, "ObjectConnectionState", group="objects", arguments=[
            UInt32Argument("object_id"),
            UInt32Argument("factory_id"),
            EnumArgument("object_type", OBJECT_TYPE, data_type=Int32Argument(), default=-1),
            BoolArgument("connected"),
        ]),
        Command(0xd1, "MotorCalibration", group="motors", arguments=[
            EnumArgument("motor_id", MOTOR_ID, data_type=UInt8Argument()),
            BoolArgument("calib_started"),
            BoolArgument("auto_started"),
        ]),
        Command(0xd7, "ObjectUpAxisChanged", group="objects", arguments=[
            UInt32Argument("timestamp"),
            UInt32Argument("object_id"),
            EnumArgument("axis", UP_AXIS, data_type=UInt8Argument(), default=7),
        ]),
        Command(0xdb, "ButtonPressed", group="system", arguments=[
            BoolArgument("pressed"),
        ]),
        Command(0xdd, "FallingStarted", group="localization", arguments=[
            UInt32Argument("unknown"),
        ]),
        Command(0xde, "FallingStopped", group="localization", arguments=[
            UInt32Argument("unknown"),
            UInt32Argument("duration_ms"),
            FloatArgument("impact_intensity"),
        ]),
        Command(0xed, "BodyInfo", group="system", arguments=[
            UInt32Argument("serial_number"),
            UInt32Argument("body_hw_version"),
            EnumArgument("body_color", BODY_COLOR, data_type=Int32Argument(), default=-1),
        ]),
        Command(0xee, "FirmwareSignature", group="system", arguments=[
            UInt16Argument("unknown"),          # Last 2 bytes of head s/n?
            StringArgument("signature"),
        ]),
        Command(0xef, "FirmwareUpdateResult", group="firmware", arguments=[
            UInt32Argument("byte_count"),
            UInt16Argument("chunk_id"),
            UInt8Argument("status"),            # 0=OK; 0x0a=complete?
        ]),

        Event(0xf0, "RobotState", group="state", arguments=[
            UInt32Argument("timestamp"),
            UInt32Argument("pose_frame_id"),
            UInt32Argument("pose_origin_id"),
            FloatArgument("pose_x"),
            FloatArgument("pose_y"),
            FloatArgument("pose_z"),
            FloatArgument("pose_angle_rad"),
            FloatArgument("pose_pitch_rad"),
            FloatArgument("lwheel_speed_mmps"),
            FloatArgument("rwheel_speed_mmps"),
            FloatArgument("head_angle_rad"),
            FloatArgument("lift_height_mm"),
            FloatArgument("accel_x"),
            FloatArgument("accel_y"),
            FloatArgument("accel_z"),
            FloatArgument("gyro_x"),
            FloatArgument("gyro_y"),
            FloatArgument("gyro_z"),
            FloatArgument("battery_voltage"),
            UInt32Argument("status"),
            FArrayArgument("cliff_data_raw", data_type=UInt16Argument(), length=4),
            UInt16Argument("backpack_touch_sensor_raw"),
            UInt8Argument("curr_path_segment"),
        ]),
        Event(0xf1, "AnimationState", group="state", arguments=[
            UInt32Argument("timestamp"),
            Int32Argument("num_anim_bytes_played"),
            Int32Argument("num_audio_frames_played"),
            UInt8Argument("enabled_anim_tracks"),
            UInt8Argument("tag"),
            UInt8Argument("client_drop_count"),
        ]),
        Event(0xf2, "ImageChunk", group="state", arguments=[
            UInt32Argument("frame_timestamp"),
            UInt32Argument("image_id"),
            UInt32Argument("chunk_debug"),
            EnumArgument("image_encoding", IMAGE_ENCODING),
            EnumArgument("image_resolution", IMAGE_RESOLUTION),
            UInt8Argument("image_chunk_count"),
            UInt8Argument("chunk_id"),
            UInt16Argument("status"),
            VArrayArgument("data"),
        ]),
        Event(0xf3, "ObjectAvailable", group="state", arguments=[
            UInt32Argument("factory_id"),
            EnumArgument("object_type", OBJECT_TYPE, data_type=Int32Argument(), default=-1),
            Int8Argument("rssi"),
        ]),
        Event(0xf4, "ImageImuData", group="state", arguments=[
            UInt32Argument("image_id"),
            FloatArgument("rate_x"),
            FloatArgument("rate_y"),
            FloatArgument("rate_z"),
            UInt8Argument("line_2_number"),
        ]),
        Event(0xf5, "ObjectAccel", group="state", arguments=[
            UInt32Argument("timestamp"),
            UInt32Argument("object_id"),
            FloatArgument("accel_x"),
            FloatArgument("accel_y"),
            FloatArgument("accel_z"),
        ]),
    ]

)




import sys
from io import StringIO

from PIL import Image


def render(image: bytes) -> None:
    for y in range(32):
        sys.stdout.write("{:3d}".format(y))
        for x in range(128):
            if ((x+1) % 10) == 1:
                sys.stdout.write(" ")
            sys.stdout.write(chr(ord("0") + image[y*128 + x])) if image[y*128 + x] else sys.stdout.write(".")
        sys.stdout.write("\n")


def image_to_str(image):
    s = StringIO()
    for y in range(32):
        for x in range(128):
            if image[y * 128 + x]:
                s.write(chr(ord("0") + image[y * 128 + x]))
            else:
                s.write(".")
        s.write("\n")
    return s.getvalue()


def str_to_image(sim: str) -> Image:
    s = StringIO(sim)
    im = Image.new("1", (128, 32), color=0)
    px = im.load()
    for y in range(32):
        while True:
            line = s.readline().strip()
            if line:
                break
        x = 0
        for c in line:
            px[x, y] = 1 if c and c != "." else 0
            x += 1
    return im


class ImageDecoder(object):

    def __init__(self, buffer: bytes):
        self.buffer = buffer
        self.image = bytearray(128 * 32)
        self.x = 0
        self.y = 0
        self.last_draw = False
        self.repeat_column_shift = False
        self.debug = False

    def _draw(self, x, y, color=1):
        assert 0 <= x < 128
        assert 0 <= y < 32
        assert 0 <= color < 256
        if y < 32 and x < 128:
            self.image[y * 128 + x] = color

    def _execute(self, b: int):
        cmd = (b & 0xc0) >> 6
        cnt = b & 0x3f
        if cmd == 0:        # Skip column
            cnt += 1
            if self.debug:
                print("Skip column {}".format(cnt))
            if self.last_draw:
                self.x += 1
            self.x += cnt
            self.y = 0
            self.last_draw = False
            self.repeat_column_shift = False
        elif cmd == 1:      # Repeat column
            cnt += 1
            if self.debug:
                print("Repeat column {}".format(cnt))
            if not self.repeat_column_shift:
                self.x += 1
            for _ in range(cnt):
                i = self.x
                for _ in range(32):
                    self.image[i] = self.image[i - 1]
                    i += 128
                self.x += 1
            self.y = 0
            self.last_draw = False
            self.repeat_column_shift = True
        elif cmd == 2:      # Draw / Skip
            draw = cnt & 0x01
            cnt >>= 1
            draw2 = cnt & 0x01
            cnt >>= 1
            cnt += 1
            if draw or draw2:
                if self.debug:
                    print("Draw {}".format(cnt))
                for _ in range(cnt):
                    self._draw(self.x, self.y)
                    self.y += 1
            else:
                if self.debug:
                    print("Skip {}".format(cnt))
                self.y += cnt
            if self.y > 31:
                self.repeat_column_shift = True
                self.x += 1
                self.y -= 32
            else:
                self.repeat_column_shift = False
            self.last_draw = True
        else:               # Draw extended / Skip extended
            draw = cnt & 0x01
            cnt >>= 1
            draw2 = cnt & 0x01
            cnt >>= 1
            cnt += 1
            cnt += 16
            if draw or draw2:
                if self.debug:
                    print("Draw extended {}".format(cnt))
                for _ in range(cnt):
                    self._draw(self.x, self.y)
                    self.y += 1
            else:
                if self.debug:
                    print("Skip extended {}".format(cnt))
                self.y += cnt
            if self.y > 31:
                self.repeat_column_shift = True
                self.x += 1
                self.y -= 32
            else:
                self.repeat_column_shift = False
            self.last_draw = False

    def decode(self) -> bytes:
        for i, b in enumerate(self.buffer):
            if self.debug:
                sys.stdout.write("{}: ".format(i))
            self._execute(b)
        return self.image


class ImageEncoder(object):

    def __init__(self, im: Image):
        if im.size[0] != 128 or im.size[1] != 32:
            raise ValueError("Invalid image dimensions. Only 128x32 images are supported.")
        if im.mode != "1":
            raise ValueError("Invalid pixel format. Only binary images are supported.")
        self.px = im.load()
        self.buffer = bytearray()
        self.last_col = bytearray()
        self.cur_col = bytearray()
        self.skip_cols = 0
        self.repeat_cols = 0
        self.x = 0
        self.y = 0

    def _encode_seq(self, color: int, cnt: int) -> int:
        """ Encode a sequence of pixels with the same color. """
        if color:
            # Draw
            if cnt <= 15:
                cmd = 0x80 + (cnt << 2) + 0x01
            else:
                assert 15 <= cnt <= 31
                cmd = 0xc0 + ((cnt - 16) << 2) + 0x01
        else:
            # Skip
            if cnt <= 15:
                cmd = 0x80 + (cnt << 2)
            elif cnt < 31:
                cmd = 0xc0 + ((cnt - 16) << 2)
            else:
                # Skip column
                assert cnt == 31
                self.skip_cols += 1
                cmd = None
        return cmd

    def _count_color(self, color: int):
        """ Count pixels with the same color, down a column. """
        cnt = 0
        if self.y < 32:
            while self.px[self.x, self.y] == color:
                cnt += 1
                self.y += 1
                if self.y > 31:
                    self.x += 1
                    self.y = 0
                    break
        else:
            self.x += 1
            self.y = 0
        return cnt

    def _skip_cols(self) -> None:
        """ Handle columns skipping. """
        if self.skip_cols:
            self.repeat_cols = 0
            self.last_col = bytearray()
            # Optimization: Remove skips, right before column skips.
            if self.buffer:
                cmd = self.buffer[-1]
                if (cmd & 0xc3 == 0x80) or (cmd & 0xc3 == 0xc0):
                    self.buffer.pop()
        while self.skip_cols >= 64:
            self.buffer.append(63)
            self.skip_cols -= 64
        if self.skip_cols:
            cmd = self.skip_cols - 1
            self.buffer.append(cmd)
            self.skip_cols = 0

    def _repeat_cols(self) -> None:
        """ Handle column repetition. """
        if self.repeat_cols:
            # Optimization: Remove skips, right before column repeats.
            if self.buffer:
                cmd = self.buffer[-1]
                if (cmd & 0xc3 == 0x80) or (cmd & 0xc3 == 0xc0):
                    self.buffer.pop()
        while self.repeat_cols >= 64:
            cmd = 0x40 + 0x3f
            self.buffer.append(cmd)
            self.repeat_cols -= 64
        if self.repeat_cols:
            cmd = 0x40 + self.repeat_cols - 1
            self.buffer.append(cmd)
            self.repeat_cols = 0

    def encode(self) -> bytearray:
        while self.x < 128 and self.y < 32:
            color = self.px[self.x, self.y]
            self.y += 1
            cnt = self._count_color(color)
            cmd = self._encode_seq(color, cnt)
            if cmd is not None:
                if self.y == 0:
                    self._skip_cols()
                    # TODO: Not clear exactly how the 2 draw bits work...
                    if (cmd & 0xc3 == 0x81) or (cmd & 0xc3 == 0xc1):
                        cmd += 1
                self.cur_col.append(cmd)
            # Handle column repetition
            if self.y == 0:
                if not self.skip_cols:
                    if self.cur_col == self.last_col:
                        self.repeat_cols += 1
                    else:
                        self._repeat_cols()
                        self.buffer.extend(self.cur_col)
                        self.last_col = self.cur_col
                else:
                    self._repeat_cols()
                self.cur_col = bytearray()
        if self.y == 0:
            self._skip_cols()
            self._repeat_cols()
        return self.buffer




from . import AnimClip
from . import AnimClips
from . import BackpackLights
from . import BodyMotion
from . import Event
from . import FaceAnimation
from . import HeadAngle
from . import Keyframes
from . import LiftHeight
from . import ProceduralFace
from . import RecordHeading
from . import RobotAudio
from . import TurnToRecordedHeading




from typing import Set


__all__ = [
    "Filter",
]


class Filter(object):

    def __init__(self):
        self.allowed_ids = set()
        self.denied_ids = set()

    def allow_ids(self, ids: Set[int]) -> None:
        self.allowed_ids.update(ids)

    def deny_ids(self, ids: Set[int]) -> None:
        self.denied_ids.update(ids)

    def filter(self, target_id: int) -> bool:
        if target_id is not None:
            if self.allowed_ids and target_id not in self.allowed_ids:
                return True
            if self.denied_ids and target_id in self.denied_ids:
                return True
        return False



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: CozmoAnim

import flatbuffers

class HeadAngle(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsHeadAngle(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = HeadAngle()
        x.Init(buf, n + offset)
        return x

    # HeadAngle
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # HeadAngle
    def TriggerTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # HeadAngle
    def DurationTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # HeadAngle
    def AngleDeg(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # HeadAngle
    def AngleVariabilityDeg(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

def HeadAngleStart(builder): builder.StartObject(4)
def HeadAngleAddTriggerTimeMs(builder, triggerTimeMs): builder.PrependUint32Slot(0, triggerTimeMs, 0)
def HeadAngleAddDurationTimeMs(builder, durationTimeMs): builder.PrependUint32Slot(1, durationTimeMs, 0)
def HeadAngleAddAngleDeg(builder, angleDeg): builder.PrependInt8Slot(2, angleDeg, 0)
def HeadAngleAddAngleVariabilityDeg(builder, angleVariabilityDeg): builder.PrependUint8Slot(3, angleVariabilityDeg, 0)
def HeadAngleEnd(builder): return builder.EndObject()




import unittest

from pycozmo.image_encoder import ImageEncoder, str_to_image, ImageDecoder, image_to_str
from pycozmo.util import hex_dump, hex_load
from pycozmo.tests.image_encoder_fixtures import FIXTURES


class TestImageEncoder(unittest.TestCase):

    @staticmethod
    def _encode(sim: str) -> str:
        im = str_to_image(sim)
        encoder = ImageEncoder(im)
        buf = encoder.encode()
        res = hex_dump(buf)
        return res

    def assertSameImage(self, sim: str, seq: str) -> None:
        buffer = hex_load(seq)
        decoder = ImageDecoder(buffer)
        decoder.decode()
        actual = image_to_str(decoder.image)
        self.assertEqual(sim.strip(), actual.strip())

    def test_blank(self):
        fixture = FIXTURES["blank"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_fill_screen(self):
        fixture = FIXTURES["fill_screen"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_fill_screen2(self):
        fixture = FIXTURES["fill_screen2"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_top_left(self):
        fixture = FIXTURES["top_left"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_top_left_5(self):
        fixture = FIXTURES["top_left_5"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_top_left_1_8(self):
        fixture = FIXTURES["top_left_1_8"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_top_left_line(self):
        fixture = FIXTURES["top_left_line"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_top_line(self):
        fixture = FIXTURES["top_line"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_bottom_line(self):
        fixture = FIXTURES["bottom_line"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_left_line(self):
        fixture = FIXTURES["left_line"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_right_line(self):
        fixture = FIXTURES["right_line"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_columns(self):
        fixture = FIXTURES["columns"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_rect(self):
        fixture = FIXTURES["rect"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_rect2(self):
        fixture = FIXTURES["rect2"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_rect3(self):
        fixture = FIXTURES["rect3"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_rect4(self):
        fixture = FIXTURES["rect4"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_diagonal(self):
        fixture = FIXTURES["diagonal"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_diagonal2(self):
        fixture = FIXTURES["diagonal2"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_blocks(self):
        fixture = FIXTURES["blocks"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_pycozmo(self):
        fixture = FIXTURES["pycozmo"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_chessboard_tl(self):
        fixture = FIXTURES["chessboard_tl"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_chessboard_bl(self):
        fixture = FIXTURES["chessboard_bl"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_chessboard_tr(self):
        fixture = FIXTURES["chessboard_tr"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_chessboard_br(self):
        fixture = FIXTURES["chessboard_br"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_chessboard2_tl(self):
        fixture = FIXTURES["chessboard2_tl"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_chessboard2_bl(self):
        fixture = FIXTURES["chessboard2_bl"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_chessboard2_tr(self):
        fixture = FIXTURES["chessboard2_tr"]
        sim = fixture["image"]
        expected = fixture["seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)

    def test_chessboard2_br(self):
        fixture = FIXTURES["chessboard2_br"]
        sim = fixture["image"]
        expected = fixture["alt_seq"]
        actual = self._encode(sim)
        self.assertEqual(expected, actual)
        self.assertSameImage(sim, actual)



#!/usr/bin/env python

import sys
import os
import dpkt
try:
    # noinspection PyPackageRequirements
    import pcap
except ImportError:
    pcap = None

import pycozmo


class DumpApp(object):

    def __init__(self):
        self.first_ts = None
        self.frame_count = 0
        self.filtered_frame_count = 0
        self.packet_count = 0
        self.unknown_count = 0
        self.filtered_packet_count = 0
        self.frame_type_filter = pycozmo.filter.Filter()
        # self.frame_type_filter.deny_ids({pycozmo.protocol_declaration.FrameType.PING.value})
        self.packet_type_filter = pycozmo.filter.Filter()
        # self.packet_type_filter.deny_ids({pycozmo.protocol_declaration.PacketType.PING.value})
        self.packet_id_filter = pycozmo.filter.Filter()
        # self.packet_id_filter.deny_ids({0x8f, 0x97})

    def decode_cozmo_frame(self, ts, buffer):
        frame = pycozmo.Frame.from_bytes(buffer)

        if self.frame_type_filter.filter(frame.type.value):
            self.filtered_frame_count += 1
            return

        print("{:<12s}first_seq={:04x}, seq={:04x}, ack={:04x}, frame={:6d}, time={:.06f}".format(
            frame.type.name, frame.first_seq, frame.seq, frame.ack, self.frame_count + 1, ts))

        for pkt in frame.pkts:
            self.packet_count += 1
            if isinstance(pkt, pycozmo.protocol_base.UnknownPacket):
                self.unknown_count += 1
            if self.packet_type_filter.filter(pkt.type.value):
                self.filtered_packet_count += 1
                continue
            if self.packet_id_filter.filter(pkt.id):
                self.filtered_packet_count += 1
                continue
            direction = "->" if pkt.is_from_robot() else "<-"
            print("\t{} time={:.06f} {}".format(direction, ts, pkt))
            # if ts > 15:
            #     sys.exit(1)

    def handle_frame(self, ts, frame):
        if self.first_ts is None:
            self.first_ts = ts
        rel_ts = ts - self.first_ts
        eth = dpkt.ethernet.Ethernet(frame)
        if eth.type != dpkt.ethernet.ETH_TYPE_IP:
            # Skip non-IP frames
            return
        ip = eth.data
        if ip.p != dpkt.ip.IP_PROTO_UDP:
            # Skip non-UDP frames
            return
        udp = ip.data
        if udp.data[:7] != pycozmo.protocol_declaration.FRAME_ID:
            # Skip non-Cozmo frames
            return
        self.decode_cozmo_frame(rel_ts, udp.data)
        self.frame_count += 1
        # if frame_count > 100:
        #     break

    def decode_pcap(self, fspec):
        with open(fspec, "rb") as f:
            for ts, frame in dpkt.pcap.Reader(f):
                self.handle_frame(ts, frame)

        print()
        print("Frames: {}".format(self.frame_count))
        print("Packets: {}".format(self.packet_count))
        if self.packet_count:
            print("Unknown packets: {} ({:.0f}%)".format(
                self.unknown_count, 100.0*self.unknown_count/self.packet_count))

    def capture(self, interface):
        pc = pcap.pcap(name=interface)
        # pc.setfilter('')
        print('listening on %s: %s' % (pc.name, pc.filter))
        pc.loop(0, self.handle_frame)


def main():
    fspec = sys.argv[1]

    app = DumpApp()

    if os.path.exists(fspec):
        app.decode_pcap(fspec)
    else:
        if pcap is None:
            sys.exit("Live packet capturing requires pypcap. Install with 'pip install pypcap'.")

        app.capture(fspec)


if __name__ == '__main__':
    main()




from threading import Event
from typing import Optional, Tuple
import json
import time
import io

import numpy as np
from PIL import Image

from . import logger
from . import protocol_encoder
from . import event
from . import camera
from . import object
from . import util
from . import robot
from . import exception
from . import filter
from . import protocol_declaration
from . import conn
from . import lights
from . import image_encoder
from . import anim
from . import anim_encoder


class Client(event.Dispatcher):

    def __init__(self, robot_addr: Optional[Tuple[str, int]] = None,
                 protocol_log_messages: Optional[list] = None) -> None:
        super().__init__()
        self.conn = conn.ClientConnection(robot_addr, protocol_log_messages)
        self.serial_number_head = None
        self.robot_fw_sig = None
        self.serial_number = None
        self.body_hw_version = None
        self.body_color = None
        # Robot state
        # Heading in X-Y plane.
        self.pose_frame_id = 0
        self.pose = util.Pose(0.0, 0.0, 0.0, angle_z=util.Angle(radians=0.0), origin_id=1)
        self.pose_pitch = util.Angle(radians=0.0)
        self.head_angle = util.Angle(radians=robot.MIN_HEAD_ANGLE.radians)
        self.left_wheel_speed = util.Speed(mmps=0.0)
        self.right_wheel_speed = util.Speed(mmps=0.0)
        self.lift_position = robot.LiftPosition(height=robot.MIN_LIFT_HEIGHT)
        self.battery_voltage = 0.0
        self.accel = util.Vector3(0.0, 0.0, 0.0)
        self.gyro = util.Vector3(0.0, 0.0, 0.0)
        self.robot_status = 0
        # Animation state
        self.num_anim_bytes_played = 0
        self.num_audio_frames_played = 0
        self.enabled_anim_tracks = 0
        self.tag = 0
        self.client_drop_count = 0
        # Camera state
        self.last_image_timestamp = None
        # Object state
        self.available_objects = dict()
        self.connected_objects = dict()
        # Filters
        self.packet_type_filter = filter.Filter()
        self.packet_type_filter.deny_ids({protocol_declaration.PacketType.PING.value})
        self.packet_id_filter = filter.Filter()
        self._reset_partial_state()
        # Animations
        self._clip_metadata = {}
        self._clips = {}
        self._ppclips = {}

    def start(self) -> None:
        logger.debug("Starting client...")
        self.conn.add_handler(protocol_encoder.HardwareInfo, self._on_hardware_info)
        self.conn.add_handler(protocol_encoder.FirmwareSignature, self._on_firmware_signature)
        self.conn.add_handler(protocol_encoder.BodyInfo, self._on_body_info)
        self.conn.add_handler(protocol_encoder.ImageChunk, self._on_image_chunk)
        self.conn.add_handler(protocol_encoder.RobotState, self._on_robot_state)
        self.conn.add_handler(protocol_encoder.AnimationState, self._on_animation_state)
        self.conn.add_handler(protocol_encoder.ObjectAvailable, self._on_object_available)
        self.conn.add_handler(protocol_encoder.ObjectConnectionState, self._on_object_connection_state)
        self.add_handler(event.EvtRobotPickedUpChange, self._on_robot_picked_up)
        self.conn.start()

    def stop(self) -> None:
        logger.debug("Stopping client...")
        self.conn.stop()
        self.del_all_handlers()

    def connect(self) -> None:
        logger.debug("Connecting...")
        self.conn.connect()

    def disconnect(self) -> None:
        logger.debug("Disconnecting...")
        self.conn.disconnect()

    def _enable_robot(self):
        # Enable
        pkt = protocol_encoder.Enable()
        self.conn.send(pkt)
        self.conn.send(pkt)  # This repetition seems to trigger BodyInfo

    def _initialize_robot(self):
        # Set world frame origin to (0,0,0), frame ID to 0, and origin ID to 1.
        pkt = protocol_encoder.SetOrigin()
        self.conn.send(pkt)
        # Set timestamp to 0. Also enables RobotState and ObjectAvailable events. Requires Enable (0x25).
        pkt = protocol_encoder.SyncTime()
        self.conn.send(pkt)
        # Enable animation playback and AnimationState events. Requires Enable (0x25).
        pkt = protocol_encoder.EnableAnimationState()
        self.conn.send(pkt)

        # Initialize display.
        for _ in range(7):
            pkt = protocol_encoder.NextFrame()
            self.conn.send(pkt)
            pkt = protocol_encoder.DisplayImage(b"\x3f\x3f")
            self.conn.send(pkt)

        # TODO: This should not be necessary.
        time.sleep(0.5)

        self.dispatch(event.EvtRobotReady, self)

    def _on_hardware_info(self, cli, pkt: protocol_encoder.HardwareInfo):
        del cli
        self.serial_number_head = pkt.serial_number_head

    def _on_firmware_signature(self, cli, pkt: protocol_encoder.FirmwareSignature):
        del cli
        self.robot_fw_sig = json.loads(pkt.signature)
        logger.info("Firmware version %s.", self.robot_fw_sig["version"])
        self._enable_robot()

    def _on_body_info(self, cli, pkt: protocol_encoder.BodyInfo):
        del cli
        self.serial_number = pkt.serial_number
        self.body_hw_version = pkt.body_hw_version
        self.body_color = pkt.body_color
        logger.info("Body S/N 0x%08x.", self.serial_number)
        supported = self.robot_fw_sig["version"] == protocol_declaration.FIRMWARE_VERSION
        if supported:
            self._initialize_robot()
        else:
            logger.error("Unsupported Cozmo firmware version %i. Only version %i is supported currently.",
                         self.robot_fw_sig["version"], protocol_declaration.FIRMWARE_VERSION)
        self.dispatch(event.EvtRobotFound, self)

    def wait_for_robot(self, timeout: float = 5.0) -> None:
        if not self.robot_fw_sig:
            e = Event()
            self.add_handler(event.EvtRobotFound, lambda cli: e.set(), one_shot=True)
            if not e.wait(timeout):
                raise exception.ConnectionTimeout("Failed to connect to Cozmo.")

        if self.robot_fw_sig["version"] != protocol_declaration.FIRMWARE_VERSION:
            raise exception.UnsupportedFirmwareVersion("Unsupported Cozmo firmware version.")

        if not self.serial_number:
            e = Event()
            self.add_handler(event.EvtRobotReady, lambda cli: e.set(), one_shot=True)
            if not e.wait(timeout):
                raise exception.ConnectionTimeout("Failed to initialize Cozmo.")

    def _reset_partial_state(self):
        self._partial_image_timestamp = None
        self._partial_data = None
        self._partial_image_id = None
        self._partial_invalid = False
        self._partial_size = 0
        self._partial_image_encoding = None
        self._partial_image_resolution = None
        self._last_chunk_id = -1

    def _on_image_chunk(self, cli, pkt: protocol_encoder.ImageChunk):
        del cli
        if self._partial_image_id is not None and pkt.chunk_id == 0:
            if not self._partial_invalid:
                logger.debug("Lost final chunk of image - discarding.")
            self._partial_image_id = None

        if self._partial_image_id is None:
            if pkt.chunk_id != 0:
                if not self._partial_invalid:
                    logger.debug("Received chunk of broken image.")
                self._partial_invalid = True
                return
            # Discard any previous in-progress image
            self._reset_partial_state()
            self._partial_image_timestamp = pkt.frame_timestamp
            self._partial_image_id = pkt.image_id
            self._partial_image_encoding = protocol_encoder.ImageEncoding(pkt.image_encoding)
            self._partial_image_resolution = protocol_encoder.ImageResolution(pkt.image_resolution)

            image_resolution = protocol_encoder.ImageResolution(pkt.image_resolution)
            width, height = camera.RESOLUTIONS[image_resolution]
            max_size = width * height * 3  # 3 bytes per pixel (RGB)
            self._partial_data = np.empty(max_size, dtype=np.uint8)

        if pkt.chunk_id != (self._last_chunk_id + 1) or pkt.image_id != self._partial_image_id:
            logger.debug("Image missing chunks - discarding (last_chunk_id=%d partial_image_id=%s).",
                         self._last_chunk_id, self._partial_image_id)
            self._reset_partial_state()
            self._partial_invalid = True
            return

        offset = self._partial_size
        self._partial_data[offset:offset + len(pkt.data)] = pkt.data
        self._partial_size += len(pkt.data)
        self._last_chunk_id = pkt.chunk_id

        if pkt.chunk_id == pkt.image_chunk_count - 1:
            self._process_completed_image()
            self._reset_partial_state()

    def _process_completed_image(self):
        data = self._partial_data[0:self._partial_size]

        # The first byte of the image is whether or not it is in color
        is_color_image = data[0] != 0

        if self._partial_image_encoding == protocol_encoder.ImageEncoding.JPEGMinimizedGray:
            width, height = camera.RESOLUTIONS[self._partial_image_resolution]

            if is_color_image:
                # Color images are half width
                width = width // 2
                data = camera.minicolor_to_jpeg(data, width, height)
            else:
                data = camera.minigray_to_jpeg(data, width, height)

        image = Image.open(io.BytesIO(data)).convert('RGB')

        # Color images need to be resized to the proper resolution
        if is_color_image:
            size = camera.RESOLUTIONS[self._partial_image_resolution]
            image = image.resize(size)

        self._latest_image = image
        self.last_image_timestamp = self._partial_image_timestamp
        self.dispatch(event.EvtNewRawCameraImage, self, image)

    def _on_robot_state(self, cli, pkt: protocol_encoder.RobotState):
        del cli
        self.pose_frame_id = pkt.pose_frame_id
        self.pose = util.Pose(pkt.pose_x, pkt.pose_y, pkt.pose_z,
                              angle_z=util.Angle(radians=pkt.pose_angle_rad), origin_id=pkt.pose_origin_id)
        self.pose_pitch = util.Angle(radians=pkt.pose_pitch_rad)
        self.head_angle = util.Angle(radians=pkt.head_angle_rad)
        self.left_wheel_speed = util.Speed(mmps=pkt.lwheel_speed_mmps)
        self.right_wheel_speed = util.Speed(mmps=pkt.rwheel_speed_mmps)
        self.lift_position = robot.LiftPosition(height=util.Distance(mm=pkt.lift_height_mm))
        self.battery_voltage = pkt.battery_voltage
        self.accel = util.Vector3(pkt.accel_x, pkt.accel_y, pkt.accel_z)
        self.gyro = util.Vector3(pkt.gyro_x, pkt.gyro_y, pkt.gyro_z)
        old_status = self.robot_status
        self.robot_status = pkt.status
        self.dispatch(event.EvtRobotStateUpdated, self)
        # Dispatch status flag change events.
        for flag, evt in event.STATUS_EVENTS.items():
            if (old_status & flag) != (pkt.status & flag):
                state = (pkt.status & flag) != 0
                logger.debug("%s: %i", robot.RobotStatusFlagNames[flag], state)
                self.dispatch(evt, self, state)

    def _on_robot_picked_up(self, cli, state):
        del cli
        if not state:
            # Robot put down - reset world frame origin.
            pkt = protocol_encoder.SetOrigin(
                pose_frame_id=self.pose_frame_id + 1, pose_origin_id=self.pose.origin_id + 1)
            self.conn.send(pkt)

    def _on_animation_state(self, cli, pkt: protocol_encoder.AnimationState):
        del cli
        self.num_anim_bytes_played = pkt.num_anim_bytes_played
        self.num_audio_frames_played = pkt.num_audio_frames_played
        self.enabled_anim_tracks = pkt.enabled_anim_tracks
        self.tag = pkt.tag
        self.client_drop_count = pkt.client_drop_count

    def _on_object_available(self, cli, pkt: protocol_encoder.ObjectAvailable):
        del cli
        factory_id = pkt.factory_id
        object_type = protocol_encoder.ObjectType(pkt.object_type)
        obj = object.Object(factory_id=factory_id, object_type=object_type)
        if factory_id not in self.available_objects:
            self.available_objects[factory_id] = obj
            logger.debug("Object of type %s with S/N 0x%08x available.", str(obj.object_type), obj.factory_id)

    def _on_object_connection_state(self, cli, pkt: protocol_encoder.ObjectConnectionState):
        del cli
        if pkt.connected:
            # Connected
            self.connected_objects[pkt.object_id] = {
                "factory_id": pkt.factory_id,
                "object_type": pkt.object_type,
            }
        else:
            # Disconnected
            if pkt.object_id in self.connected_objects:
                del self.connected_objects[pkt.object_id]

    def set_head_angle(self, angle: float, accel: float = 10.0, max_speed: float = 10.0,
                       duration: float = 0.0):
        pkt = protocol_encoder.SetHeadAngle(angle_rad=angle, accel_rad_per_sec2=accel,
                                            max_speed_rad_per_sec=max_speed, duration_sec=duration)
        self.conn.send(pkt)

    def move_head(self, speed: float) -> None:
        pkt = protocol_encoder.MoveHead(speed_rad_per_sec=speed)
        self.conn.send(pkt)

    def set_lift_height(self, height: float, accel: float = 10.0, max_speed: float = 10.0,
                        duration: float = 0.0):
        pkt = protocol_encoder.SetLiftHeight(height_mm=height, accel_rad_per_sec2=accel,
                                             max_speed_rad_per_sec=max_speed, duration_sec=duration)
        self.conn.send(pkt)

    def move_lift(self, speed: float) -> None:
        pkt = protocol_encoder.MoveLift(speed_rad_per_sec=speed)
        self.conn.send(pkt)

    def drive_wheels(self, lwheel_speed: float, rwheel_speed: float,
                     lwheel_acc: Optional[float] = 0.0, rwheel_acc: Optional[float] = 0.0,
                     duration: Optional[float] = None) -> None:
        pkt = protocol_encoder.DriveWheels(lwheel_speed_mmps=lwheel_speed, rwheel_speed_mmps=rwheel_speed,
                                           lwheel_accel_mmps2=lwheel_acc, rwheel_accel_mmps2=rwheel_acc)
        self.conn.send(pkt)
        if duration is not None:
            time.sleep(duration)
            self.stop_all_motors()

    def stop_all_motors(self) -> None:
        pkt = protocol_encoder.StopAllMotors()
        self.conn.send(pkt)

    def go_to_pose(self, pose: util.Pose, relative_to_robot: bool = False) -> None:
        """ Move to a specific pose (position and orientation). """

        if relative_to_robot:
            pose = util.Pose(self.pose.position.x, self.pose.position.y,
                             self.pose.position.z, angle_z=self.pose.rotation.angle_z).define_pose_relative_this(pose)

        pkt = protocol_encoder.AppendPathSegLine(
            from_x=self.pose.position.x, from_y=self.pose.position.y,
            to_x=pose.position.x, to_y=pose.position.y,
            speed_mmps=100.0, accel_mmps2=20.0, decel_mmps2=20.0)
        self.conn.send(pkt)
        pkt = protocol_encoder.AppendPathSegPointTurn(
            x=pose.position.x, y=pose.position.y,
            angle_rad=pose.rotation.angle_z.radians,
            angle_tolerance_rad=0.01,
            speed_mmps=40.0, accel_mmps2=20.0, decel_mmps2=20.0)
        self.conn.send(pkt)
        pkt = protocol_encoder.ExecutePath(event_id=1)
        self.conn.send(pkt)

        e = Event()

        def event_wait(_, pkt2: protocol_encoder.PathFollowingEvent):
            if pkt2.event_type != protocol_encoder.PathEventType.PATH_STARTED:
                e.set()

        self.conn.add_handler(protocol_encoder.PathFollowingEvent, event_wait)
        e.wait()

    def set_backpack_lights(self, left_light, front_light, center_light, rear_light, right_light) -> None:
        pkt = protocol_encoder.LightStateCenter(states=(front_light, center_light, rear_light))
        self.conn.send(pkt)
        pkt = protocol_encoder.LightStateSide(states=(left_light, right_light))
        self.conn.send(pkt)

    def set_center_backpack_lights(self, light) -> None:
        self.set_backpack_lights(lights.off_light, light, light, light, lights.off_light)

    def set_all_backpack_lights(self, light) -> None:
        self.set_backpack_lights(light, light, light, light, light)

    def set_backpack_lights_off(self) -> None:
        self.set_backpack_lights(lights.off_light, lights.off_light, lights.off_light,
                                 lights.off_light, lights.off_light)

    def set_head_light(self, enable: bool) -> None:
        pkt = protocol_encoder.SetHeadLight(enable=enable)
        self.conn.send(pkt)

    def display_image(self, im: Image, duration: Optional[float] = None) -> None:
        encoder = image_encoder.ImageEncoder(im)
        buf = bytes(encoder.encode())
        pkt = protocol_encoder.DisplayImage(image=buf)
        self.conn.send(pkt)
        self.next_frame()
        if duration is not None:
            time.sleep(duration)

    def next_frame(self):
        pkt = protocol_encoder.NextFrame()
        self.conn.send(pkt)

    def _load_clips(self, fspec: str) -> None:
        if fspec.endswith(".bin"):
            clips = anim_encoder.AnimClips.from_fb_file(fspec)
        elif fspec.endswith(".json"):
            clips = anim_encoder.AnimClips.from_json_file(fspec)
        else:
            raise ValueError("Unsupported animation file format.")
        for clip in clips.clips:
            self._clips[clip.name] = clip

    def play_anim(self, name: str) -> None:
        if not self._clip_metadata:
            raise ValueError("Animations not loaded.")
        elif name not in self._clip_metadata:
            raise ValueError("Unknown clip name.")

        if name not in self._ppclips:
            if name not in self._clips:
                self._load_clips(self._clip_metadata[name].fspec)
            clip = self._clips[name]
            self._ppclips[name] = anim.PreprocessedClip.from_anim_clip(clip)

        ppclip = self._ppclips[name]
        ppclip.play(self)

    def load_anims(self, dspec: str) -> None:
        self._clip_metadata = anim_encoder.get_clip_metadata(dspec)
        self._clips = {}

    def get_anim_names(self) -> set:
        return set(self._clip_metadata.keys())

    @property
    def anim_names(self) -> set:
        return self.get_anim_names()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: CozmoAnim

import flatbuffers

class RecordHeading(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRecordHeading(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RecordHeading()
        x.Init(buf, n + offset)
        return x

    # RecordHeading
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RecordHeading
    def TriggerTimeMs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

def RecordHeadingStart(builder): builder.StartObject(1)
def RecordHeadingAddTriggerTimeMs(builder, triggerTimeMs): builder.PrependUint32Slot(0, triggerTimeMs, 0)
def RecordHeadingEnd(builder): return builder.EndObject()
